import os
import asyncio
import threading
import time
import signal
import re

from typing import Dict, List, Optional, Tuple
from pathlib import Path

from rfilerunner.colors import Colors, color
from rfilerunner.util import (
    verbose,
    error,
    ngather,
    VERBOSE,
)
from rfilerunner.parse import Params
from rfilerunner import runners

import watchdog
import watchdog.observers

# List of running process IDs (used to terminate running processes from
# another thread in the event of a watch cancel)
_procs = {}


def strip_ansi(s: str) -> str:
    ansi_escape = re.compile(
        r"""
        \x1B  # ESC
        (?:   # 7-bit C1 Fe (except CSI)
            [@-Z\\-_]
        |     # or [ for CSI, followed by a control sequence
            \[
            [0-?]*  # Parameter bytes
            [ -/]*  # Intermediate bytes
            [@-~]   # Final byte
        )
    """,
        re.VERBOSE,
    )

    return ansi_escape.sub("", s)


def isfloat(s: str) -> bool:
    try:
        float(s)
        return True
    except ValueError:
        return False


def observer_join(observer):
    try:
        observer.join()
    except KeyboardInterrupt:
        print("overserver")
        return


async def aiojoin(observer):
    # Wrapper to join via asyncio
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, observer_join, observer)


def worker(loop):
    # Worker thread to execute an async loop (filled with tasks by a file
    # listener)
    asyncio.set_event_loop(loop)
    loop.run_forever()


class Handler(watchdog.events.FileSystemEventHandler):
    def __init__(self, procs, params, watch_run):
        self.last_handle = None
        self.loop = asyncio.new_event_loop()
        self.watch_run = watch_run
        self.procs = procs
        self.params = params

        worker_thread = threading.Thread(target=worker, args=(self.loop,))
        worker_thread.start()

    def on_any_event(self, event):
        # On a file change event, run the relevant script
        if self.params.cancel_watch:

            # Terminate any async processes in the worker thread's loop
            if self.last_handle is not None:
                self.last_handle.cancel()

            verbose(event)

            # Stop any previous runs of this command if present (this gets
            # filled out in runners.py when the process is actually run via
            # subprocess)
            if self.procs[self.params.name] is not None:
                try:
                    os.kill(self.procs[self.params.name], signal.SIGKILL)
                except ProcessLookupError:
                    pass

            self.last_handle = asyncio.run_coroutine_threadsafe(
                self.watch_run(event), self.loop
            )
        else:
            asyncio.run(self.watch_run(event))


async def watch(
    params: Params,
    args: Dict[str, str],
    commands: Dict[str, Params],
    cwd: str,
    padding: int,
    run_idx: Optional[int],
    watch_files: Optional[List[str]] = None,
):
    """
    Run a command via a watch. This loops infinitely based on 'params' (e.g.
    based on files or a timer).
    """

    async def catch(rc, stdout):
        pass

    if params.catch is not None:
        if params.catch in commands:
            dependency_params = commands[params.catch]

            async def catch(rc, stdout):
                new_args = args.copy()
                new_args["ERROR"] = strip_ansi(stdout)
                new_args["ERROR_COLOR"] = stdout
                rc, stdout = await run(
                    dependency_params,
                    new_args,
                    commands,
                    cwd,
                )

        else:

            async def catch(rc, stdout):
                new_args = args.copy()
                new_args["ERROR"] = strip_ansi(stdout)
                new_args["ERROR_COLOR"] = stdout
                run_code = params.catch + "\n"
                run_params = Params(
                    name=f"{params.name}-catch",
                    shell=params.shell,
                    help="",
                    args=new_args,
                    deps=[],
                    parallel=False,
                    watch=None,
                    catch=None,
                    code=run_code,
                    cancel_watch=False,
                )
                rc, stdout = await runners.shell(run_params, new_args, cwd)

    async def watch_run(event):
        new_args = args.copy()

        if event is not None:
            new_args["CHANGED"] = event.src_path

        rc, stdout = await runners.shell(
            params,
            new_args,
            cwd,
            padding=padding,
            run_idx=run_idx,
            running_pids=_procs[params.name],
        )
        if rc != 0:
            await catch(rc, stdout)

    _procs[params.name] = None
    handler = Handler(_procs, params, watch_run)
    if run_idx is None:
        # no prefix if this isn't run alongside other commands
        preamble = ""
    else:
        # prepend with: "<name> |"
        preamble = f"{params.name}{' ' * (padding - len(params.name))} | "

    if watch_files is not None:
        # files passed in via CLI override, don't run anything
        paths_to_watch = [Path(x) for x in watch_files]
    else:
        if params.watch in commands:
            # Recurse down to another run to get its output
            dependency_params = commands[params.watch]
            rc, stdout = await run(
                dependency_params,
                args,
                commands,
                cwd,
                padding=padding,
                run_idx=run_idx,
                hide_output=True,
            )
        elif isfloat(params.watch):
            # Passed a timer so no need to listen to any files, just sit here
            # in a loop forever
            sleep_time = float(params.watch)
            print(
                color(
                    f"{preamble}watching every {sleep_time} seconds",
                    Colors.YELLOW,
                )
            )
            while True:
                await watch_run(None)
                time.sleep(sleep_time)
        else:
            # Inline shell code used, so issue an ad-hoc call to the default
            # shell
            new_args = params.args.copy()
            new_args["CHANGED"] = ""
            run_code = params.watch + "\n"
            run_params = Params(
                name=f"{params.name}-watch",
                shell=params.shell,
                help="",
                args=new_args,
                deps=[],
                parallel=False,
                watch=None,
                catch=None,
                code=run_code,
                cancel_watch=False,
            )
            rc, stdout = await runners.shell(
                run_params, new_args, cwd, hide_output=True
            )
            # This is probably unreachable
            if rc != 0:
                error(f"watch command failed: {run_code.strip()}\n{stdout.rstrip()}")

        paths_to_watch = [
            Path(x.strip()) for x in stdout.split("\n") if x.strip() != ""
        ]

    # Check that any files referenced exist (watchdog will error out otherwise
    # with a less helpful message)
    non_existent = [p for p in paths_to_watch if not p.exists()]
    if len(non_existent):
        non_existent = ", ".join([str(x) for x in non_existent])
        error(f"Some paths to watch didn't exist: {non_existent}")

    # Output watch status
    paths_str = " ".join([str(x) for x in paths_to_watch])
    if len(paths_str) > 140 and not VERBOSE:
        print(color(f"{preamble}watching {len(paths_to_watch)} files", Colors.YELLOW))
    else:
        print(
            color(
                f"{preamble}watching {' '.join([str(x) for x in paths_to_watch])}",
                Colors.YELLOW,
            )
        )

    # Register all the relevant paths with watchdog
    observer = watchdog.observers.Observer()

    for path in paths_to_watch:
        # TODO: inotify limits seem pretty low (around 100?), so watching a whole
        # repo doesn't work
        observer.schedule(handler, str(path.resolve()), recursive=False)

    observer.start()

    # Run once to start
    if params.cancel_watch:
        handler.on_any_event(None)
    else:
        await watch_run(None)

    # This should loop forever
    await aiojoin(observer)

    return 0, None


async def run(
    params: Params,
    args: Dict[str, str],
    commands: Dict[str, Params],
    cwd: str,
    padding: int = 0,
    run_idx: Optional[int] = None,
    hide_output: bool = False,
    no_watch: bool = False,
    no_parallel: bool = False,
    watch_files: Optional[List[str]] = None,
) -> Tuple[int, str]:
    """
    Execute an rfile command and any transitive dependencies.
    """
    verbose(f"Running command {params.name}, {params}")

    # Run dependencies
    if len(params.deps) > 0:
        # Actual invocations don't know about the others, so compute the padding
        # for each output line and pass it down
        padding = max(len(p) for p in params.deps)

        for dep in params.deps:
            if dep not in commands:
                error(
                    f"'{dep}' command not found in rfile but was specified as a dependency of '{params.name}'"
                )

        # Gather a set of coroutines for each dependency. These all run in the
        # same async loop but yield for reading from stdout of each subprocess.
        dependency_runs = [
            run(
                commands[dep],
                args,
                commands,
                cwd,
                padding=padding,
                run_idx=i,
                hide_output=hide_output,
            )
            for i, dep in enumerate(params.deps)
        ]

        if not no_parallel and params.parallel:
            await ngather(dependency_runs)
        else:
            for c in dependency_runs:
                await c

    if params.code.strip() == "":
        # No actual code (but can't do this any earlier in case there are dependencies)
        return 0, ""

    if params.watch is not None and not no_watch:
        # This shouldn't ever actually return, just spin forever watching the
        # specified files
        return await watch(
            params, args, commands, cwd, padding, run_idx, watch_files=watch_files
        )
    else:
        # Normal run, determine the runner based on params.shell
        runner = runners.generic
        if params.shell.name in {"bash", "zsh", "sh", "fish"}:
            runner = runners.shell
        elif params.shell.name in {"python", "python3"}:
            runner = runners.python

        # Execute code
        rc, stdout = await runner(
            params, args, cwd, padding=padding, run_idx=run_idx, hide_output=hide_output
        )
        return rc, stdout



import subprocess
import sys
import json
import asyncio
import os
import multiprocessing
from typing import Any, Optional

VERBOSE = False
from rfilerunner.colors import Colors, color


NPROC = multiprocessing.cpu_count()
c = Colors
usable_colors = [c.BLUE, c.PURPLE, c.GREEN, c.YELLOW, c.LIGHT_RED, c.LIGHT_GREEN]


def cmd(s, **kwargs):
    print(s)
    subprocess.run(s, shell=True, **kwargs)


def error(error_message, code=1):
    print(f"{color('User error:', Colors.RED)} {error_message}", file=sys.stderr)
    os._exit(code)


def verbose(message):
    if VERBOSE:
        print(message)


def check(cond, message):
    if not cond:
        error(message)


def internal_assert(cond, message):
    if not cond:
        print(message)
        error("If you're seeing this, this is an rfile bug")


def color_from_run(run_idx: Optional[int]) -> str:
    if run_idx is None:
        return ""

    return usable_colors[run_idx % len(usable_colors)]


def padding_from_run(name, run_info):
    padding = 0
    if run_info is not None and "padding" in run_info:
        padding = run_info["padding"] - len(name)

    return " " * padding


async def ngather(tasks, n=NPROC - 1):
    semaphore = asyncio.Semaphore(n)

    async def sem_task(task):
        async with semaphore:
            return await task

    return await asyncio.gather(*(sem_task(task) for task in tasks))


def jprint(o):
    """
    JSON dump an object
    """
    print(json.dumps(o, indent=2, default=str))



import shutil
from typing import Dict, List, Optional, Tuple
from typing import NamedTuple
from pathlib import Path
from rfilerunner.util import (
    error,
    check,
)
from rfilerunner.colors import Colors, color


class Params(NamedTuple):
    shell: str
    name: str
    args: Dict[str, str]  # name: arg(help, value)
    help: str
    watch: str
    catch: str
    deps: List[str]
    parallel: bool
    code: str
    cancel_watch: bool


def default_shell() -> Optional[str]:
    """
    Find a shell on the system via shutil
    """
    items = ["bash", "zsh", "sh"]
    for i in items:
        path = shutil.which(i)
        if path is not None:
            return path

    error(f"No shell found, tried: {items}")


def parse_name_and_help(s: str) -> Tuple[str, str]:
    """
    handle stuff like
    # shell: /bin/sh (my help text)
    """
    parts = s.split(" ")
    name = parts[0]
    rest = None
    if len(parts) > 1:
        rest = " ".join(parts[1:])
        rest = rest.lstrip("(").rstrip(")")

    return name, rest


def parse(name: str, code: str, is_default: bool) -> Params:
    """
    Parse an entry in an rfile (a single command, i.e. a key + string in the
    top level dictionary in the rfile).

    This handles these directives (see docs for details):
    # shell
    # parallel
    # dep
    # help
    # watch
    # arg
    """
    lines = code.split("\n")
    preamble = []

    for index, line in enumerate(lines):
        line = line.strip()
        if line.startswith("# "):
            preamble.append(line)
        else:
            # TODO: Probably broken if it's all preamble
            break

    code = lines[index:]

    # Set these all to defaults, figure them out if present in parsing
    shell = default_shell()
    help = None
    args = {}
    deps = []
    cancel_watch = False
    parallel = False
    catch = None
    watch_target = None

    for line in preamble:
        if line.startswith("# shell: "):
            shell_path, help = parse_name_and_help(line[len("# shell: ") :])
            shell = shutil.which(shell_path)
            check(shell is not None, f"Shell {shell_path} could not be found in PATH")
        elif line.startswith("# arg: "):
            arg, arg_help = parse_name_and_help(line[len("# arg: ") :])
            args[arg] = arg_help if arg_help is not None else ""
        elif line.startswith("# dep: "):
            deps.append(line[len("# dep: ") :])
        elif line.startswith("# help: "):
            help = line[len("# help: ") :]
        elif line.strip() == "# parallel":
            parallel = True
        elif line.startswith("# watch:"):
            watch_target = line[len("# watch:") :].strip()
        elif line.startswith("# catch:"):
            catch = line[len("# catch:") :].strip()
        elif line.strip() == "# cancel":
            cancel_watch = True
        elif help is None:
            help = line[len("# ") :]

    # If this has no help but dependencies, infer a help based on the dependencies
    if help is None and len(deps) > 0:
        if len(deps) == 1:
            help = f"run {deps[0]}"
        if len(deps) == 2:
            help = f"run {deps[0]} and {deps[1]}"
        else:
            help = "run " + ", ".join(deps[:-1]) + ", and " + deps[-1]

    full_code = "\n".join(code)

    if help is None:
        # If help is still blank, put in some help text so it's not blank,
        # just use the code
        sample = full_code[:31].strip().replace("\n", "; ")
        if len(sample) == 31:
            sample = sample[:27] + "..."
        else:
            sample = sample[:30]
        help = color(sample, Colors.FAINT)

    # Label the first command
    if is_default:
        if help is None:
            help = "(default)"
        else:
            help += " (default)"

    return Params(
        name=name,
        shell=Path(shell),
        help=help,
        args=args,
        deps=deps,
        parallel=parallel,
        watch=watch_target,
        catch=catch,
        code=full_code,
        cancel_watch=cancel_watch,
    )



import subprocess
import tempfile
import os
import sys
import textwrap
import pty
import asyncio
from typing import Dict, Optional, List, Set

from rfilerunner import util
from rfilerunner import colors, Colors
from rfilerunner.parse import Params
from rfilerunner.util import verbose, padding_from_run, color_from_run

c = Colors


def read(f):
    return os.read(f, 1000)


async def aioread(f):
    """
    asyncio wrapper for os.read
    detials: https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
    """
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, read, f)


async def run_in_interpreter(
    params: Params,
    args: Dict[str, str],
    cwd: str,
    preamble: str,
    running_pids: Optional[Set[int]] = None,
    hide_output: bool = False,
    padding: int = 0,
    run_idx: Optional[int] = None,
):
    """
    Actually execute 'params'
    """
    verbose(f"Shell executing: {params}")
    env = os.environ.copy()
    for k, v in args.items():
        env[k.lower()] = v
        env[k.upper()] = v

    single = run_idx is None
    recorded_stdout = ""

    with tempfile.NamedTemporaryFile(delete=False) as f:
        with open(f.name, "w") as f_w:
            f_w.write(preamble)
            f_w.write(params.code)

        # Use a pseudo-terminal so subprocesses send color output (color codes
        # will be stripped out later if necessary)
        pty_r, pty_w = pty.openpty()
        stderr = pty_w
        stdout = pty_w

        command = [params.shell, f_w.name] + list(args.values())
        command_args = [f_w.name] + list(args.values())
        verbose(f"  running {command}")

        # Calculate padding based on other runs
        padding = padding - len(params.name)
        proc = await asyncio.create_subprocess_exec(
            params.shell,
            *command_args,
            cwd=cwd,
            env=env,
            stderr=stderr,
            stdout=stdout,
        )

        # Store the PID in case this needs to be terminated for a cancel by the
        # file event listener
        if running_pids is not None:
            running_pids.add(proc.pid)

        padding = " " * padding
        color = ""
        if run_idx is not None:
            color = util.usable_colors[run_idx % len(util.usable_colors)]

        os.close(pty_w)

        # Read from the pty output until the process finishes
        while True:
            try:
                output = await aioread(pty_r)
            except OSError as e:
                verbose(f"OSError: {e}")
                break
            if not output:
                # subprocess closed stdout / stderr
                break

            output = output.decode()
            if not hide_output:
                if single:
                    print(output, end="")
                else:
                    lines = output.split("\n")
                    end = len(lines)
                    if output.endswith("\n"):
                        end = end - 1
                    for line in lines[:end]:
                        print(f"{color}{params.name}{c.END}{padding} | {line.rstrip()}")

            recorded_stdout += output
            sys.stdout.flush()

        aioout, aioerr = await proc.communicate()

        # All output should already have been read in the loop above
        if aioout is not None:
            print(f"Unexpected output after closing read pipe:\n{aioout}")
        if aioerr is not None:
            print(f"Unexpected output after closing read pipe:\n{aioerr}")

        return proc.returncode, recorded_stdout


def python(params, args, cwd, **kwargs):
    # Build args to pass into Python as a dotdict (defined below)
    arg_data = {}
    for name in params.args:
        arg_data[name] = None
    for name, value in args.items():
        arg_data[name] = f'"{value}"'

    data = [f"{k}={v}" for k, v in arg_data.items()]
    data = ", ".join(data)

    # Import some common stuff to avoid boilerplate in scripts
    preamble = textwrap.dedent(
        f"""
    import os
    import sys
    import subprocess
    import math
    import re
    import json
    import random

    class dotdict(dict):
        __getattr__ = dict.get
        __setattr__ = dict.__setitem__
        __delattr__ = dict.__delitem__

    args = dotdict({data})
    """
    )
    return run_in_interpreter(params, args, cwd, preamble=preamble, **kwargs)


def shell(params, args, cwd, **kwargs):
    # It's a shell, so error out when a command fails via set -e
    preamble = "set -e\n"
    if util.VERBOSE:
        preamble = "set -ex\n"
    return run_in_interpreter(params, args, cwd, preamble=preamble, **kwargs)


def generic(params, args, cwd, **kwargs):
    # Execution environment isn't known, so add nothing and just pass through
    return run_in_interpreter(params, args, cwd, preamble="", **kwargs)



import setuptools
import subprocess


with open("README.md", "r") as f:
    long_description = f.read()


def fetch_tag():
    proc = subprocess.run(
        ["git", "tag", "--points-at", "HEAD"],
        stdout=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    )
    tags = proc.stdout.strip().split("\n")
    if tags[0] == "":
        raise RuntimeError("no tags found")
    return tags[0].replace("v", "")


def shorthash():
    proc = subprocess.run(
        ["git", "log", "-1", "--format='%h'"],
        stdout=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    )
    return proc.stdout.strip().replace("'", "")


try:
    version = fetch_tag()
except Exception as e:
    print("Failed to fetch tag", e)
    try:
        version = f"git+{shorthash()}"
    except Exception as e:
        print("Failed to fetch shorthash", e)
        version = "0.1.4"

print("VERSION", version)
setuptools.setup(
    name="rfilerunner",
    version=version,
    description="a simple command runner",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/goatmobile/rfilerunner",
    author="goatmobile",
    entry_points={
        "console_scripts": ["r = rfilerunner:cli", "rfile = rfilerunner:cli"]
    },
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    install_requires=["Click", "PyYAML", "watchdog"],
)



import sys
import os
import textwrap
import argparse
import yaml
import asyncio
import signal

from pathlib import Path
from typing import Any


from rfilerunner import util
from rfilerunner.colors import Colors, color
from rfilerunner.util import (
    verbose,
    error,
    internal_assert,
    check,
)
from rfilerunner.parse import parse
from rfilerunner.run import run


def check_rfile(content: Any):
    """
    Check that an rfile's content is the right structure
    """
    check(
        isinstance(content, dict),
        f"Expected rfile top level be a flat YAML dictionary, but found {type(content)}",
    )

    for v in content.values():
        check(
            isinstance(v, str),
            f"Expected rfile dictionary entries to be strings, but found {type(v)}",
        )


def show_help(missing_file: bool, commands, error=None):
    preamble = textwrap.dedent(
        f"""
    {color('usage', Colors.YELLOW)}: r [-h, --help] [-v, --verbose] [-r, --rfile rfile] {color('COMMAND', Colors.CYAN)}

    rfile is a simple command runner for executing Python and shell scripts
    """
    ).strip()

    if missing_file:
        print(
            preamble
            + "\n\n"
            + f"{color('note', Colors.GREEN)}: no rfile was found so no commands are available"
        )
    else:
        command_str = ""
        max_len = max(len(name) for name in commands.keys())

        for name, params in commands.items():
            line = f"    {color(name, Colors.PURPLE)}"
            padding = " " * (max_len - len(name) + 5)
            if params.help is not None:
                line += f"{padding}{params.help}"

            command_str += f"{line}\n"
        print(
            f"{preamble}\n\n{color('available commands:', Colors.BOLD)}\n{command_str}",
            end="",
        )

    if error is not None:
        print(f"{Colors.BOLD}{Colors.RED}{error}{Colors.END}")

    exit(0)


def show_subcommand_help(command, params):
    preamble = (
        f"{color('usage', Colors.YELLOW)}: r [-v, --verbose] {command} [-h, --help]"
    )
    args_desc = ""
    args = [f"[--{a} {a.upper()}]" for a in params.args]
    if len(args) > 0:
        preamble += " " + " ".join(args)

    arg_strs = [("  -h, --help", "[r] show this help message and exit")]
    if params.watch:
        arg_strs.append(
            (
                "  --no-watch / --once",
                "[r] disable '# watch' behavior and only run once",
            )
        )
        arg_strs.append(
            (
                "  --watch WATCH",
                "[r] comma-separated list of files to watch (override '# watch')",
            )
        )
    if params.parallel:
        arg_strs.append(
            (
                "  --no-parallel / --serial",
                "[r] disable 'parallel' behavior and run dependencies serially",
            )
        )

    for name, help in params.args.items():
        arg_strs.append((f"  --{name} {name.upper()}", help))
    padding = max(len(x) for x, _ in arg_strs)
    args_descs = [
        f"{name}{' ' * (padding - len(name) + 2)}{help}" for name, help in arg_strs
    ]
    args_desc = f"\n\n{color('optional arguments:', Colors.BOLD)}\n" + "\n".join(
        args_descs
    )
    print(
        f"{preamble}\n\n    {params.help if len(params.help) > 0 else '(no script found in command)'}{args_desc}"
    )
    exit(0)


def signal_handler(signal, frame):
    os._exit(0)


signal.signal(signal.SIGINT, signal_handler)


def locate_rfile(help: bool) -> str:
    rfile_names = ["rfile", "rfile.yml", "rfile.yaml"]
    verbose(f"Guessing rfile from {rfile_names}")
    cwd = Path(os.getcwd())

    for part in [cwd] + list(cwd.parents):
        possible_rfiles = [part / f for f in rfile_names]
        existing_rfiles = [f for f in possible_rfiles if f.exists() and f.is_file()]
        if len(existing_rfiles) == 1:
            return existing_rfiles[0]

    sample = textwrap.dedent(
        """
    # rfiles are just yaml, but comments under commands have meaning

    my_prereq1:
        echo working...

    my_prereq2:
        echo really working...

    my_command:
        # arg: name (your name)
        # parallel
        # dep: my_prereq1
        # dep: my_prereq2
        echo hello "$NAME"
    """.rstrip()
    )
    msg = color(
        "File 'rfile' not found in this directory or parents! Make one! Here is an example:",
        Colors.RED,
    )
    error(msg + sample)


def handle_shell_completions(prev, options):
    shell = os.getenv("SHELL", None)
    if shell is None:
        print("Didn't find anything in SHELL environment variable, exiting")
        exit(0)
    shell = Path(shell)

    if prev is not None:
        prev = prev.split()
    else:
        prev = []

    if shell.name == "fish":
        if hasattr(sys.stdout, "isatty") and sys.stdout.isatty():
            completions_dir = (
                Path(os.path.expanduser("~")) / ".config" / "fish" / "completions"
            )
            out = completions_dir / "r.fish"

            print(f"Install completions for fish shell in {out}?\n(y/n) ", end="")
            response = input().lower()
            data = textwrap.dedent(
                """
                function get_r_completions
                    set prev_arg (commandline -pco)
                    r --completions --prev "$prev_arg"
                end

                complete -c r -k -a '(get_r_completions)' --no-files
            """
            )
            if response == "y":
                completions_dir.mkdir(exist_ok=True, parents=True)
                with open(out, "w") as f:
                    f.write(data)
                print("Installed successfully!")

            else:
                printable = data.replace("'", "\\'")
                print(
                    f"'no' chosen, aborting. You can install manually via:\necho '{printable}' > {out}"
                )
        else:
            if len(prev) == 1:
                # print out all the commands if none has been entered yet
                for command, params in options.items():
                    print(f"{command}\t{params.help}")
            elif len(prev) == 2:
                if prev[-1] == "-r" or prev[-1] == "--rfile":
                    # re-implement file listing since that's what is needed here
                    for item in Path(".").glob("*"):
                        if item.is_file() and item.suffix in {".yaml", ".yml"}:
                            print(item)
                else:
                    # options specific to a command, find it and print
                    curr_command = prev[-1]
                    if curr_command in options:
                        params = options[curr_command]
                        for arg, help in params.args.items():
                            print(f"--{arg}\t{help}")
    else:
        print(f"Shell '{shell.name}' isn't supported, only these shells are: fish")

    exit(0)


def cli():
    # First argparser to silently grab the rfile and verbose args before
    # building the real one with user supplied info
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-r", "--rfile")
    parser.add_argument("-h", "--help", action="store_true")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "--completions",
        action="store_true",
        help="(dev) Print shell completions",
    )
    parser.add_argument(
        "--prev",
        help="(dev) Previous shell args",
        default=None,
    )
    parser.add_argument("args", nargs=argparse.REMAINDER)
    args, other = parser.parse_known_args()

    first_help = args.help

    if args.verbose:
        util.VERBOSE = True

    if args.rfile:
        verbose(f"Using hardcoded rfile at {args.rfile}")
        rfile = Path(args.rfile)
        if not rfile.exists():
            if args.help:
                show_help(missing_file=True, commands=None)
            error(f"Could not find rfile '{args.rfile}'")
    else:
        rfile = locate_rfile(help=args.help)

    # Load the file
    internal_assert(rfile.exists(), "check rfile exists")

    with open(rfile) as f:
        content = yaml.safe_load(f)
    check_rfile(content)

    # Generate the real args
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-r",
        "--rfile",
        help="the YAML rfile to use ('rfile' in the current directory by default)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose output for debugging",
    )
    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
    )
    parser.add_argument("subparser", nargs="?")

    default = None
    commands = {}
    for index, (command, code) in enumerate(content.items()):
        is_default = index == 0
        if is_default:
            default = command
        params = parse(name=command, code=code, is_default=is_default)

        commands[params.name] = params

    if args.completions:
        handle_shell_completions(args.prev, commands)

    check(default is not None, "There should be at least 1 command in the rfile")
    args, other = parser.parse_known_args()

    command = args.subparser
    if args.subparser is None:
        command = default
    if command not in commands:
        # If the commmand isn't a literal name of an command, try to guess it
        # based on the prefix
        candidates = [item for item in commands if item.startswith(command)]
        if len(candidates) == 0:
            # No matches
            show_help(
                missing_file=False,
                commands=commands,
                error=f"No possible matches found for command '{command}'",
            )
        elif len(candidates) > 1:
            # Too many matches
            candidates = ", ".join(candidates)
            show_help(
                missing_file=False,
                commands=commands,
                error=f"Ambiguous short command '{command}', could be any of: {candidates}",
            )
        else:
            # Just right, use the full command
            print(
                color(
                    f"Assuming '{command}' is short for '{candidates[0]}'",
                    Colors.YELLOW,
                )
            )
            command = candidates[0]

    inner_parser = argparse.ArgumentParser(add_help=False)
    params = commands[command]
    for name, help in params.args.items():
        inner_parser.add_argument(f"--{name}")

    if params.watch:
        inner_parser.add_argument("--no-watch", action="store_true")
        inner_parser.add_argument("--once", action="store_true")
        inner_parser.add_argument("--watch")

    if params.parallel:
        inner_parser.add_argument("--no-parallel", action="store_true")
        inner_parser.add_argument("--serial", action="store_true")

    subparser_args, subparser_other = inner_parser.parse_known_args()

    if first_help or args.help:
        if args.subparser is None:
            show_help(missing_file=False, commands=commands)
        else:
            show_subcommand_help(command, commands[command])

    params = commands[command]
    runtime_args = {}

    for name in params.args.keys():
        value = getattr(subparser_args, name, None)
        if value is not None:
            runtime_args[name] = value

    no_watch = False
    if params.watch and (subparser_args.no_watch or subparser_args.once):
        no_watch = True

    no_parallel = False
    if params.parallel and (subparser_args.no_parallel or subparser_args.serial):
        no_parallel = True

    watch_files = None
    if params.watch and subparser_args.watch:
        if no_watch:
            error("Cannot use --watch with --no-watch/--once")

        watch_files = [x.strip() for x in subparser_args.watch.split(",")]

    code, stdout = asyncio.run(
        run(
            params,
            runtime_args,
            commands,
            cwd=rfile.parent,
            no_watch=no_watch,
            no_parallel=no_parallel,
            watch_files=watch_files,
        )
    )
    exit(code)
